import json


from paignion.exception import PaignionException


class PaignionRoom(object):
    def __init__(
        self,
        name,
        description,
        north=None,
        east=None,
        south=None,
        west=None,
        up=None,
        down=None,
        tangible_items=None,
        intangible_items=None,
    ):
        self.name = name
        self.description = description
        self.north = north
        self.east = east
        self.south = south
        self.west = west
        self.up = up
        self.down = down
        self.tangible_items = tangible_items
        self.intangible_items = intangible_items

        self.verify_attributes()

    def verify_attributes(self):
        # Tangible items should be an empty list by default
        self.tangible_items = [] if not self.tangible_items else self.tangible_items
        # Intangible items should be an empty list by default
        self.intangible_items = (
            [] if not self.intangible_items else self.intangible_items
        )

        # Item name is mandatory
        if not self.name:
            raise PaignionException("Name missing for item")

        # Item description is mandatory
        if not self.description:
            raise PaignionException(f"Description missing for item `{self.name}`")

    def dump(self):
        return {
            self.name: {
                "north": self.north,
                "east": self.east,
                "south": self.south,
                "west": self.west,
                "up": self.up,
                "down": self.down,
                "description": self.description,
                "items": {
                    "tangible": [i.dump() for i in self.tangible_items],
                    "intangible": [i.dump() for i in self.intangible_items],
                },
            }
        }

    def __str__(self):
        return json.dumps(self.dump(), indent=4)



from paignion.exception import PaignionException
from paignion.action_compiler import ActionCompiler


class PaignionUsedWithItem(object):
    def __init__(
        self,
        name,
        effect_message,
        consumes_subject=True,
        consumes_object=True,
        actions=None,
    ):
        self.name = name
        self.effect_message = effect_message
        self.consumes_subject = consumes_subject
        self.consumes_object = consumes_object
        self.actions = actions

        self.verify_attributes()

    def verify_attributes(self):
        # Subject should not be consumed by default
        self.consumes_subject = False if not self.consumes_subject else True
        # Object should not be consumed by default
        self.consumes_object = False if not self.consumes_object else True

        # Item name is mandatory
        if not self.name:
            raise PaignionException("Name missing for used_with item")

        # Effect message is mandatory
        if not self.effect_message:
            raise PaignionException(
                f"Effect message missing for used_with item `{self.name}`"
            )

        action_compiler = ActionCompiler()
        self.actions = "".join(
            [action_compiler.compile_action(a) for a in self.actions]
        )

    def dump(self):
        return {
            "name": self.name,
            "effect_message": self.effect_message,
            "consumes_subject": self.consumes_subject,
            "consumes_object": self.consumes_object,
            "actions": self.actions,
        }

    def __str__(self):
        return json.dumps(self.dump(), indent=4)



import setuptools

from paignion.definitions import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="paignion",
    version=__version__,
    author="Dimitri Kokkonis",
    author_email="kokkonisd@gmail.com",
    description="A simple game engine for adventure text games.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kokkonisd/paignion",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=["pyyaml", "markdown", "pymdown-extensions"],
    package_data={"paignion": ["frontend/*"]},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "paignion = paignion.__main__:paignion_main_function",
        ],
    },
    python_requires=">=3.6",
)



import re

from paignion.exception import PaignionException
from paignion.tools import markdownify


class ActionToken(object):
    def __init__(self, type, value):
        self.type = type
        self.value = value

    def __str__(self):
        return f"{self.type.upper()}_TOKEN (`{self.value}`)"

    def __repr__(self):
        return self.__str__()


class ActionSetNode(object):
    def __init__(self, key, value, element):
        self.key = key
        self.value = value
        self.element = element

    def action(self):
        return "set"

    def __str__(self):
        return f"{self.action().upper()}_NODE({self.key}, {self.value}, {self.element})"


class ActionAddNode(ActionSetNode):
    def action(self):
        return "add"


class ActionSubNode(ActionSetNode):
    def action(self):
        return "sub"


class ActionMulNode(ActionSetNode):
    def action(self):
        return "mul"


class ActionDivNode(ActionSetNode):
    def action(self):
        return "div"


class ActionCompiler(object):
    TOKEN_TYPES = [
        ("set_func", r"set"),
        ("add_func", r"add"),
        ("sub_func", r"sub"),
        ("mul_func", r"mul"),
        ("div_func", r"div"),
        ("md_string", r'm"(([^"\\])|(\\"))*"'),
        ("string", r'"(([^"\\])|(\\"))*"'),
        ("integer", r"[0-9]+"),
        ("identifier", r"[a-zA-Z_][a-zA-Z0-9_]*"),
        ("oparen", r"\("),
        ("cparen", r"\)"),
        ("comma", r","),
        ("whitespace", r"[ \t\r]"),
    ]

    def consume_token(self, action):
        for token in self.TOKEN_TYPES:
            # Attempt to match action
            match = re.match(token[1], action)
            if match:
                # Consume token
                action = action[match.end(0) :]
                return ActionToken(token[0], match[0]), action

        return None, action

    def tokenize_action(self, action, filter_out=None):
        token_list = []
        original_action = action

        while action:
            token, action = self.consume_token(action)

            if not token:
                raise PaignionException(
                    f"Undefined token `{action[0]}` in `{action}`, in action"
                    f" `{original_action}`"
                )

            if token.type not in filter_out:
                token_list.append(token)

        return token_list

    def parse_action(self, action):
        original_action = action
        token_list = self.tokenize_action(action, filter_out=["whitespace"])

        while token_list:
            if self.peek("set_func", token_list):
                return self.parse_set_func(token_list)
            elif self.peek("add_func", token_list):
                return self.parse_add_func(token_list)
            elif self.peek("sub_func", token_list):
                return self.parse_sub_func(token_list)
            elif self.peek("mul_func", token_list):
                return self.parse_mul_func(token_list)
            elif self.peek("div_func", token_list):
                return self.parse_div_func(token_list)

            raise PaignionException(
                f"Undefined structure starting with `{token_list[0]}` for action"
                f" `{original_action}`"
            )

    def compile_action(self, action):
        action_node = self.parse_action(action)

        if action_node.action() == "set":
            return (
                f'getRoomOrItem("{action_node.element}")["{action_node.key}"] = '
                f"{action_node.value};"
            )
        elif action_node.action() == "add":
            return (
                f'getRoomOrItem("{action_node.element}")["{action_node.key}"] += '
                f"{action_node.value};"
            )
        elif action_node.action() == "sub":
            return (
                f'getRoomOrItem("{action_node.element}")["{action_node.key}"] -= '
                f"{action_node.value};"
            )
        elif action_node.action() == "mul":
            return (
                f'getRoomOrItem("{action_node.element}")["{action_node.key}"] *= '
                f"{action_node.value};"
            )
        elif action_node.action() == "div":
            return (
                f'getRoomOrItem("{action_node.element}")["{action_node.key}"] /= '
                f"{action_node.value};"
            )

        raise PaignionException(f"Undefined node: `{action_node}`")

    def parse_set_func(self, token_list):
        self.consume("set_func", token_list)
        self.consume("oparen", token_list)

        # Key, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            key = self.parse_string(token_list)
        else:
            key = self.parse_identifier(token_list)

        self.consume("comma", token_list)

        # Value, can be integer, Markdown string or simple string
        if self.peek("integer", token_list):
            value = self.parse_integer(token_list)
        elif self.peek("md_string", token_list):
            value = self.parse_md_string(token_list)
        else:
            value = self.parse_string(token_list)

        self.consume("comma", token_list)

        # Element, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            element = self.parse_string(token_list)
        else:
            element = self.parse_identifier(token_list)

        self.consume("cparen", token_list)

        return ActionSetNode(key=key, value=value, element=element)

    def parse_add_func(self, token_list):
        self.consume("add_func", token_list)
        self.consume("oparen", token_list)

        # Value, must be either integer, Markdown string or string
        if self.peek("integer", token_list):
            value = self.parse_integer(token_list)
        elif self.peek("md_string", token_list):
            value = self.parse_md_string(token_list)
        else:
            value = self.parse_string(token_list)

        self.consume("comma", token_list)

        # Key, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            key = self.parse_string(token_list)
        else:
            key = self.parse_identifier(token_list)

        self.consume("comma", token_list)

        # Element, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            element = self.parse_string(token_list)
        else:
            element = self.parse_identifier(token_list)

        self.consume("cparen", token_list)

        return ActionAddNode(key=key, value=value, element=element)

    def parse_sub_func(self, token_list):
        self.consume("sub_func", token_list)
        self.consume("oparen", token_list)

        # Value, must be integer
        value = self.parse_integer(token_list)

        self.consume("comma", token_list)

        # Key, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            key = self.parse_string(token_list)
        else:
            key = self.parse_identifier(token_list)

        self.consume("comma", token_list)

        # Element, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            element = self.parse_string(token_list)
        else:
            element = self.parse_identifier(token_list)

        self.consume("cparen", token_list)

        return ActionSubNode(key=key, value=value, element=element)

    def parse_mul_func(self, token_list):
        self.consume("mul_func", token_list)
        self.consume("oparen", token_list)

        # Value, must be integer
        value = self.parse_integer(token_list)

        self.consume("comma", token_list)

        # Key, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            key = self.parse_string(token_list)
        else:
            key = self.parse_identifier(token_list)

        self.consume("comma", token_list)

        # Element, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            element = self.parse_string(token_list)
        else:
            element = self.parse_identifier(token_list)

        self.consume("cparen", token_list)

        return ActionMulNode(key=key, value=value, element=element)

    def parse_div_func(self, token_list):
        self.consume("div_func", token_list)
        self.consume("oparen", token_list)

        # Value, must be integer
        value = self.parse_integer(token_list)

        if value == 0:
            raise PaignionException("div() call with 0 detected, cannot divide by 0")

        self.consume("comma", token_list)

        # Key, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            key = self.parse_string(token_list)
        else:
            key = self.parse_identifier(token_list)

        self.consume("comma", token_list)

        # Element, can be string or identifier (based on if it has spaces or not)
        if self.peek("string", token_list):
            element = self.parse_string(token_list)
        else:
            element = self.parse_identifier(token_list)

        self.consume("cparen", token_list)

        return ActionDivNode(key=key, value=value, element=element)

    def parse_md_string(self, token_list):
        # Keep the quotes!
        return f'"{markdownify(self.consume("md_string", token_list).value[2:-1])}"'

    def parse_string(self, token_list):
        # Keep the quotes!
        return f'"{self.consume("string", token_list).value[1:-1]}"'

    def parse_identifier(self, token_list):
        return str(self.consume("identifier", token_list).value)

    def parse_integer(self, token_list):
        return int(self.consume("int", token_list).value)

    def consume(self, expected_type, token_list):
        token = token_list.pop(0)

        if token.type == expected_type:
            return token

        raise PaignionException(
            f"Expected token type `{expected_type}` but got `{token.type}`"
        )

    def peek(self, expected_type, token_list, index=0):
        return token_list[index].type == expected_type



import os


__version__ = "0.0.5"

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
FRONTEND_DIR = os.path.join(BASE_DIR, "frontend")


DIRECTIONS = [
    "north",
    "east",
    "south",
    "west",
    "up",
    "down",
]

MD_EXTENSIONS = [
    "pymdownx.tilde",
    "pymdownx.emoji",
]


SIMPLE_ORIGIN_ROOM_TEMPLATE = """\
---
## Paignion v{}
##
## This is a template for a basic origin room in Paignion.
## There must always be exactly one origin room in a Paignion game; this is the room
## where the player starts in when they launch the game.
##
## Room files, as you can see, are split into two parts: a YAML header and a Markdown
## body. The YAML head is enclosed in three dashes (-) and must always be present, even
## if it is empty. In the YAML part you can define a room's metadata; the rooms it
## links to, the items it contains, the properties of these items etc. The Markdown
## body contains the description of the room, which the player sees whenever they enter
## it.
##
## As you can see below, you do not need to define all directions, you just need to
## define the ones you are going to use. In this example, at the east of this room we
## find the second room of the game, which means that if the player writes `go east`
## from this room then they will move to that second room. The possible directions are
## north, east, south, west, up and down. You can not define some or all directions and
## fill them in dynamically via actions later (see the `complete_demo` example game and
## the README for more info).
east: second_room
items:
    intangible:
        - name: painting
          description: A gorgeous painting.
    tangible:
        - name: book
          description: An old, dusty book.
---

This is the start room. There is a painting on the wall and a book on the _floor_.
"""


SIMPLE_SECOND_ROOM_TEMPLATE = """\
---
## Paignion v{}
##
## This is a template for a basic second room in Paignion.
west: origin
---

This is the second room. Not much going on here.
"""
